from django.http import HttpResponse
from django.template import Context, Template, engines
from django.test import RequestFactory

from csp.middleware import CSPMiddleware


def response(*args, headers=None, **kwargs):
    def get_response(req):
        response = HttpResponse(*args, **kwargs)
        if headers:
            for k, v in headers.items():
                response.headers[k] = v
        return response

    return get_response


JINJA_ENV = engines["jinja2"]
mw = CSPMiddleware(response())
rf = RequestFactory()


class ScriptTestBase:
    def assert_template_eq(self, tpl1, tpl2):
        aaa = tpl1.replace("\n", "").replace("  ", "")
        bbb = tpl2.replace("\n", "").replace("  ", "")
        assert aaa == bbb, f"{aaa} != {bbb}"

    def process_templates(self, tpl, expected):
        request = rf.get("/")
        mw.process_request(request)
        ctx = self.make_context(request)
        return (
            self.make_template(tpl).render(ctx).strip(),
            expected.format(request.csp_nonce),
        )


class ScriptTagTestBase(ScriptTestBase):
    def make_context(self, request):
        return Context({"request": request})

    def make_template(self, tpl):
        return Template(tpl)


class ScriptExtensionTestBase(ScriptTestBase):
    def make_context(self, request):
        return {"request": request}

    def make_template(self, tpl):
        return JINJA_ENV.from_string(tpl)



from django.conf import settings
from django.test.utils import override_settings
from django.utils.functional import lazy

from csp.utils import build_policy


def policy_eq(a, b, msg="%r != %r"):
    parts_a = sorted(a.split("; "))
    parts_b = sorted(b.split("; "))
    assert parts_a == parts_b, msg % (a, b)


def test_empty_policy():
    policy = build_policy()
    assert "default-src 'self'" == policy


def literal(s):
    return s


lazy_literal = lazy(literal, str)


@override_settings(CSP_DEFAULT_SRC=["example.com", "example2.com"])
def test_default_src():
    policy = build_policy()
    assert "default-src example.com example2.com" == policy


@override_settings(CSP_SCRIPT_SRC=["example.com"])
def test_script_src():
    policy = build_policy()
    policy_eq("default-src 'self'; script-src example.com", policy)


@override_settings(CSP_SCRIPT_SRC_ATTR=["example.com"])
def test_script_src_attr():
    policy = build_policy()
    policy_eq("default-src 'self'; script-src-attr example.com", policy)


@override_settings(CSP_SCRIPT_SRC_ELEM=["example.com"])
def test_script_src_elem():
    policy = build_policy()
    policy_eq("default-src 'self'; script-src-elem example.com", policy)


@override_settings(CSP_OBJECT_SRC=["example.com"])
def test_object_src():
    policy = build_policy()
    policy_eq("default-src 'self'; object-src example.com", policy)


@override_settings(CSP_PREFETCH_SRC=["example.com"])
def test_prefetch_src():
    policy = build_policy()
    policy_eq("default-src 'self'; prefetch-src example.com", policy)


@override_settings(CSP_STYLE_SRC=["example.com"])
def test_style_src():
    policy = build_policy()
    policy_eq("default-src 'self'; style-src example.com", policy)


@override_settings(CSP_STYLE_SRC_ATTR=["example.com"])
def test_style_src_attr():
    policy = build_policy()
    policy_eq("default-src 'self'; style-src-attr example.com", policy)


@override_settings(CSP_STYLE_SRC_ELEM=["example.com"])
def test_style_src_elem():
    policy = build_policy()
    policy_eq("default-src 'self'; style-src-elem example.com", policy)


@override_settings(CSP_IMG_SRC=["example.com"])
def test_img_src():
    policy = build_policy()
    policy_eq("default-src 'self'; img-src example.com", policy)


@override_settings(CSP_MEDIA_SRC=["example.com"])
def test_media_src():
    policy = build_policy()
    policy_eq("default-src 'self'; media-src example.com", policy)


@override_settings(CSP_FRAME_SRC=["example.com"])
def test_frame_src():
    policy = build_policy()
    policy_eq("default-src 'self'; frame-src example.com", policy)


@override_settings(CSP_FONT_SRC=["example.com"])
def test_font_src():
    policy = build_policy()
    policy_eq("default-src 'self'; font-src example.com", policy)


@override_settings(CSP_CONNECT_SRC=["example.com"])
def test_connect_src():
    policy = build_policy()
    policy_eq("default-src 'self'; connect-src example.com", policy)


@override_settings(CSP_SANDBOX=["allow-scripts"])
def test_sandbox():
    policy = build_policy()
    policy_eq("default-src 'self'; sandbox allow-scripts", policy)


@override_settings(CSP_SANDBOX=[])
def test_sandbox_empty():
    policy = build_policy()
    policy_eq("default-src 'self'; sandbox", policy)


@override_settings(CSP_REPORT_URI="/foo")
def test_report_uri():
    policy = build_policy()
    policy_eq("default-src 'self'; report-uri /foo", policy)


@override_settings(CSP_REPORT_URI=lazy_literal("/foo"))
def test_report_uri_lazy():
    policy = build_policy()
    policy_eq("default-src 'self'; report-uri /foo", policy)


@override_settings(CSP_REPORT_TO="some_endpoint")
def test_report_to():
    policy = build_policy()
    policy_eq("default-src 'self'; report-to some_endpoint", policy)


@override_settings(CSP_IMG_SRC=["example.com"])
def test_update_img():
    policy = build_policy(update={"img-src": "example2.com"})
    policy_eq("default-src 'self'; img-src example.com example2.com", policy)


def test_update_missing_setting():
    """update should work even if the setting is not defined."""
    policy = build_policy(update={"img-src": "example.com"})
    policy_eq("default-src 'self'; img-src example.com", policy)


@override_settings(CSP_IMG_SRC=["example.com"])
def test_replace_img():
    policy = build_policy(replace={"img-src": "example2.com"})
    policy_eq("default-src 'self'; img-src example2.com", policy)


def test_replace_missing_setting():
    """replace should work even if the setting is not defined."""
    policy = build_policy(replace={"img-src": "example.com"})
    policy_eq("default-src 'self'; img-src example.com", policy)


def test_config():
    policy = build_policy(config={"default-src": ["'none'"], "img-src": ["'self'"]})
    policy_eq("default-src 'none'; img-src 'self'", policy)


@override_settings(CSP_IMG_SRC=("example.com",))
def test_update_string():
    """
    GitHub issue #40 - given project settings as a tuple, and
    an update/replace with a string, concatenate correctly.
    """
    policy = build_policy(update={"img-src": "example2.com"})
    policy_eq("default-src 'self'; img-src example.com example2.com", policy)


@override_settings(CSP_IMG_SRC=("example.com",))
def test_replace_string():
    """
    Demonstrate that GitHub issue #40 doesn't affect replacements
    """
    policy = build_policy(replace={"img-src": "example2.com"})
    policy_eq("default-src 'self'; img-src example2.com", policy)


@override_settings(CSP_FORM_ACTION=["example.com"])
def test_form_action():
    policy = build_policy()
    policy_eq("default-src 'self'; form-action example.com", policy)


@override_settings(CSP_BASE_URI=["example.com"])
def test_base_uri():
    policy = build_policy()
    policy_eq("default-src 'self'; base-uri example.com", policy)


@override_settings(CSP_CHILD_SRC=["example.com"])
def test_child_src():
    policy = build_policy()
    policy_eq("default-src 'self'; child-src example.com", policy)


@override_settings(CSP_FRAME_ANCESTORS=["example.com"])
def test_frame_ancestors():
    policy = build_policy()
    policy_eq("default-src 'self'; frame-ancestors example.com", policy)


@override_settings(CSP_NAVIGATE_TO=["example.com"])
def test_navigate_to():
    policy = build_policy()
    policy_eq("default-src 'self'; navigate-to example.com", policy)


@override_settings(CSP_MANIFEST_SRC=["example.com"])
def test_manifest_src():
    policy = build_policy()
    policy_eq("default-src 'self'; manifest-src example.com", policy)


@override_settings(CSP_WORKER_SRC=["example.com"])
def test_worker_src():
    policy = build_policy()
    policy_eq("default-src 'self'; worker-src example.com", policy)


@override_settings(CSP_PLUGIN_TYPES=["application/pdf"])
def test_plugin_types():
    policy = build_policy()
    policy_eq("default-src 'self'; plugin-types application/pdf", policy)


@override_settings(CSP_REQUIRE_SRI_FOR=["script"])
def test_require_sri_for():
    policy = build_policy()
    policy_eq("default-src 'self'; require-sri-for script", policy)


@override_settings(CSP_REQUIRE_TRUSTED_TYPES_FOR=["'script'"])
def test_require_trusted_types_for():
    policy = build_policy()
    policy_eq("default-src 'self'; require-trusted-types-for 'script'", policy)


@override_settings(CSP_TRUSTED_TYPES=["strictPolicy", "laxPolicy", "default", "'allow-duplicates'"])
def test_trusted_types():
    policy = build_policy()
    policy_eq(
        "default-src 'self'; trusted-types strictPolicy laxPolicy default 'allow-duplicates'",
        policy,
    )


@override_settings(CSP_UPGRADE_INSECURE_REQUESTS=True)
def test_upgrade_insecure_requests():
    policy = build_policy()
    policy_eq("default-src 'self'; upgrade-insecure-requests", policy)


@override_settings(CSP_BLOCK_ALL_MIXED_CONTENT=True)
def test_block_all_mixed_content():
    policy = build_policy()
    policy_eq("default-src 'self'; block-all-mixed-content", policy)


def test_nonce():
    policy = build_policy(nonce="abc123")
    policy_eq("default-src 'self' 'nonce-abc123'", policy)


@override_settings(CSP_INCLUDE_NONCE_IN=["script-src", "style-src"])
def test_nonce_include_in():
    policy = build_policy(nonce="abc123")
    policy_eq(
        "default-src 'self'; script-src 'nonce-abc123'; style-src 'nonce-abc123'",
        policy,
    )


@override_settings()
def test_nonce_include_in_absent():
    del settings.CSP_INCLUDE_NONCE_IN
    policy = build_policy(nonce="abc123")
    policy_eq("default-src 'self' 'nonce-abc123'", policy)



from functools import wraps


def csp_exempt(f):
    @wraps(f)
    def _wrapped(*a, **kw):
        r = f(*a, **kw)
        r._csp_exempt = True
        return r

    return _wrapped


def csp_update(**kwargs):
    update = {k.lower().replace("_", "-"): v for k, v in kwargs.items()}

    def decorator(f):
        @wraps(f)
        def _wrapped(*a, **kw):
            r = f(*a, **kw)
            r._csp_update = update
            return r

        return _wrapped

    return decorator


def csp_replace(**kwargs):
    replace = {k.lower().replace("_", "-"): v for k, v in kwargs.items()}

    def decorator(f):
        @wraps(f)
        def _wrapped(*a, **kw):
            r = f(*a, **kw)
            r._csp_replace = replace
            return r

        return _wrapped

    return decorator


def csp(**kwargs):
    config = {k.lower().replace("_", "-"): [v] if isinstance(v, str) else v for k, v in kwargs.items()}

    def decorator(f):
        @wraps(f)
        def _wrapped(*a, **kw):
            r = f(*a, **kw)
            r._csp_config = config
            return r

        return _wrapped

    return decorator



from django.http import (
    HttpResponse,
    HttpResponseNotFound,
    HttpResponseServerError,
)
from django.test import RequestFactory
from django.test.utils import override_settings

from csp.middleware import CSPMiddleware
from csp.tests.utils import response

HEADER = "Content-Security-Policy"
mw = CSPMiddleware(response())
rf = RequestFactory()


def test_add_header():
    request = rf.get("/")
    response = HttpResponse()
    mw.process_response(request, response)
    assert HEADER in response


def test_exempt():
    request = rf.get("/")
    response = HttpResponse()
    response._csp_exempt = True
    mw.process_response(request, response)
    assert HEADER not in response


@override_settings(CSP_EXCLUDE_URL_PREFIXES=("/inlines-r-us"))
def text_exclude():
    request = rf.get("/inlines-r-us/foo")
    response = HttpResponse()
    mw.process_response(request, response)
    assert HEADER not in response


@override_settings(CSP_REPORT_ONLY=True)
def test_report_only():
    request = rf.get("/")
    response = HttpResponse()
    mw.process_response(request, response)
    assert HEADER not in response
    assert HEADER + "-Report-Only" in response


def test_dont_replace():
    request = rf.get("/")
    response = HttpResponse()
    response[HEADER] = "default-src example.com"
    mw.process_response(request, response)
    assert response[HEADER] == "default-src example.com"


def test_use_config():
    request = rf.get("/")
    response = HttpResponse()
    response._csp_config = {"default-src": ["example.com"]}
    mw.process_response(request, response)
    assert response[HEADER] == "default-src example.com"


def test_use_update():
    request = rf.get("/")
    response = HttpResponse()
    response._csp_update = {"default-src": ["example.com"]}
    mw.process_response(request, response)
    assert response[HEADER] == "default-src 'self' example.com"


@override_settings(CSP_IMG_SRC=["foo.com"])
def test_use_replace():
    request = rf.get("/")
    response = HttpResponse()
    response._csp_replace = {"img-src": ["bar.com"]}
    mw.process_response(request, response)
    policy_list = sorted(response[HEADER].split("; "))
    assert policy_list == ["default-src 'self'", "img-src bar.com"]


@override_settings(DEBUG=True)
def test_debug_errors_exempt():
    request = rf.get("/")
    response = HttpResponseServerError()
    mw.process_response(request, response)
    assert HEADER not in response


@override_settings(DEBUG=True)
def test_debug_notfound_exempt():
    request = rf.get("/")
    response = HttpResponseNotFound()
    mw.process_response(request, response)
    assert HEADER not in response


def test_nonce_created_when_accessed():
    request = rf.get("/")
    mw.process_request(request)
    nonce = str(request.csp_nonce)
    response = HttpResponse()
    mw.process_response(request, response)
    assert nonce in response[HEADER]


def test_no_nonce_when_not_accessed():
    request = rf.get("/")
    mw.process_request(request)
    response = HttpResponse()
    mw.process_response(request, response)
    assert "nonce-" not in response[HEADER]


def test_nonce_regenerated_on_new_request():
    request1 = rf.get("/")
    request2 = rf.get("/")
    mw.process_request(request1)
    mw.process_request(request2)
    nonce1 = str(request1.csp_nonce)
    nonce2 = str(request2.csp_nonce)
    assert request1.csp_nonce != request2.csp_nonce

    response1 = HttpResponse()
    response2 = HttpResponse()
    mw.process_response(request1, response1)
    mw.process_response(request2, response2)
    assert nonce1 not in response2[HEADER]
    assert nonce2 not in response1[HEADER]


@override_settings(CSP_INCLUDE_NONCE_IN=[])
def test_no_nonce_when_disabled_by_settings():
    request = rf.get("/")
    mw.process_request(request)
    nonce = str(request.csp_nonce)
    response = HttpResponse()
    mw.process_response(request, response)
    assert nonce not in response[HEADER]



import base64
import http.client as http_client
import os
from functools import partial

from django.conf import settings
from django.utils.deprecation import MiddlewareMixin
from django.utils.functional import SimpleLazyObject

from csp.utils import build_policy


class CSPMiddleware(MiddlewareMixin):
    """
    Implements the Content-Security-Policy response header, which
    conforming user-agents can use to restrict the permitted sources
    of various content.

    See http://www.w3.org/TR/CSP/

    """

    def _make_nonce(self, request):
        # Ensure that any subsequent calls to request.csp_nonce return the
        # same value
        if not getattr(request, "_csp_nonce", None):
            request._csp_nonce = base64.b64encode(os.urandom(16)).decode("ascii")
        return request._csp_nonce

    def process_request(self, request):
        nonce = partial(self._make_nonce, request)
        request.csp_nonce = SimpleLazyObject(nonce)

    def process_response(self, request, response):
        if getattr(response, "_csp_exempt", False):
            return response

        # Check for ignored path prefix.
        prefixes = getattr(settings, "CSP_EXCLUDE_URL_PREFIXES", ())
        if request.path_info.startswith(prefixes):
            return response

        # Check for debug view
        status_code = response.status_code
        exempted_debug_codes = (
            http_client.INTERNAL_SERVER_ERROR,
            http_client.NOT_FOUND,
        )
        if status_code in exempted_debug_codes and settings.DEBUG:
            return response

        header = "Content-Security-Policy"
        if getattr(settings, "CSP_REPORT_ONLY", False):
            header += "-Report-Only"

        if header in response:
            # Don't overwrite existing headers.
            return response

        response[header] = self.build_policy(request, response)

        return response

    def build_policy(self, request, response):
        config = getattr(response, "_csp_config", None)
        update = getattr(response, "_csp_update", None)
        replace = getattr(response, "_csp_replace", None)
        nonce = getattr(request, "_csp_nonce", None)
        return build_policy(config=config, update=update, replace=replace, nonce=nonce)
